import json
import time
import numpy as np
import ftlock
from http import HTTPStatus
from txtapi.models.counter import Counter
from txtapi.models.input import Input
from txtapi.txt.txt_state import TxtState
from txtapi.utility.watch_process import WatchProcess
from txtapi.txt.txt import Txt
from txtapi.handlers import util_handler
from fischertechnik.controller.Motor import Motor

# flags for event streams
is_input_message_stream_open = False
is_camera_message_stream_open = False
is_counter_message_stream_open = False
is_application_message_stream_open = False

def init_controller(controller_id):
    try:
        Txt.get_instance().init_controller(controller_id)
    except Exception as ex:
        return None, HTTPStatus.NOT_FOUND
    return None, HTTPStatus.OK


def create_counter(controller_id, counter_id):
    Txt.get_instance().create_counter(controller_id, counter_id)
    return None, HTTPStatus.OK


def create_counters(controller_id, counters):
    for idx in range(len(counters)):
        create_counter(controller_id, idx + 1)
    return None, HTTPStatus.OK


def create_input(controller_id, input_id, input_device):
    Txt.get_instance().create_input(controller_id, input_id, input_device)
    return None, HTTPStatus.OK


def create_inputs(controller_id, input_devices):
    for idx in range(len(input_devices)):
        create_input(controller_id, idx + 1, input_devices[idx])
    return None, HTTPStatus.OK


def set_motor(controller_id, motor_id, motor_options):
    motor = Txt.get_instance().get_motor(controller_id, motor_id)
    if motor is not None:
        speed = motor_options.values[0] 
        direction = Motor.CCW
        if motor_options.direction == "CW":
            direction = Motor.CW
        motor.set_speed(speed, direction)
        motor.start()
    return None, HTTPStatus.OK


def set_servomotor(controller_id, servomotor_id, servomotor_options):
    servomotor = Txt.get_instance().get_servomotor(controller_id, servomotor_id)
    if servomotor is not None:
        servomotor.set_position(servomotor_options.value)
    return None, HTTPStatus.OK


def set_output(controller_id, output_id, output_options):
    output = Txt.get_instance().get_output(controller_id, output_id)
    if output is not None:
        output.instance.setLevel(output_options.values[(output_id - 1) % 2])
    return None, HTTPStatus.OK


def get_input(controller_id, input_id):
    response = Input()
    response.name = 'I' + str(input_id)
    input = Txt.get_instance().get_input(controller_id, input_id)
    if input is not None and hasattr(input, 'get_distance'):
        response.value = input.get_distance()
    elif input is not None and hasattr(input, 'get_resistance'):
        response.value = input.get_resistance()
    elif input is not None and hasattr(input , 'get_voltage'):
        response.value = input.get_voltage()
    elif input is not None and hasattr(input , 'get_state'):
        response.value = input.get_state()
    else:
        response.value = None
    return response, HTTPStatus.OK


def get_inputs(controller_id):
    response = []
    inputs = Txt.get_instance().get_inputs(controller_id)
    if inputs is not None:
        for idx in range(len(inputs)):
            res, _ = get_input(controller_id, idx + 1)
            response.append(res)
    return response, HTTPStatus.OK


def get_counter(controller_id, counter_id):
    response = Counter()
    response.name = 'C' + str(counter_id)
    counter = Txt.get_instance().get_counter(controller_id, counter_id)
    if counter is not None:
        response.state = counter.get_state()
        response.count = counter.get_count()
    else:
        response.state = 0
        response.count = 0
    return response, HTTPStatus.OK


def get_counters(controller_id):
    response = []
    counters = Txt.get_instance().get_counters(controller_id)
    if counters is not None:
        for idx in range(len(counters)):
            res, _ = get_counter(controller_id, idx + 1)
            response.append(res)
    return response, HTTPStatus.OK


def reset_counter(controller_id, counter_id):
    counter = Txt.get_instance().get_counter(controller_id, counter_id)
    if counter is not None:
        counter.reset()
    return None, HTTPStatus.OK


def add_camera_image_recognition_config(config):
    Txt.get_instance().add_camera_image_recognition_config(config)
    return None, HTTPStatus.OK


def start_camera(config):
    try:
        util_handler.stop()
    except:
        pass
    Txt.get_instance().start_camera(config)
    return None, HTTPStatus.OK


def stop_camera():
    Txt.get_instance().stop_camera()
    return None, HTTPStatus.OK


def camera_image_stream():
    camera = Txt.get_instance().get_camera()
    while camera.is_running() == True:
        success, frame, _ = camera.read()
        if success and frame is not None:
            image = camera.frame_to_jpg(frame, 30)
            # yield the output frame in the byte format
            yield (b'--frame\r\n' b'Content-Type: image/jpeg\r\n\r\n' + bytearray(image) + b'\r\n')


def camera_message_stream():

    # stop all camera message streams before starting a new one    
    global is_camera_message_stream_open
    is_camera_message_stream_open = False
    time.sleep(1.0)
    is_camera_message_stream_open = True

    previous_output = []
    camera = Txt.get_instance().get_camera()
    start = int(time.time())
    while is_camera_message_stream_open:
        success, _, output = camera.read()
        if success and output is not None and not np.array_equal(output, previous_output):
            previous_output = output
            for item in output:
                yield 'data: %s\n\n' % json.dumps(item)
        else:
            # send an empty message after 45 seconds to avoid breaking the connection   
            end = int(time.time())
            if end - start >= 45:
                start = end
                yield 'data: []\n\n'
        # wait 50 ms for next iteration
        time.sleep(0.05)
        
    print("camera message stream closed")


def application_message_stream():

    # stop all application message streams before starting a new one
    global is_application_message_stream_open
    is_application_message_stream_open = False
    time.sleep(1.0)
    is_application_message_stream_open = True

    start = int(time.time())
    while is_application_message_stream_open:
        messages = []
        while WatchProcess.get_instance().has_messages():
            messages.append(WatchProcess.get_instance().get_message(True, 1))
        if messages:
            yield 'data: %s\n\n' % json.dumps(messages)
        else:
            # send an empty message after 45 seconds to avoid breaking the connection   
            end = int(time.time())
            if end - start >= 45:
                start = end
                yield 'data: []\n\n'
        # wait 50 ms for next iteration
        time.sleep(0.05)

    print("application message stream closed")


def input_message_stream(controller_id):

    # stop all input message streams before starting a new one
    global is_input_message_stream_open
    is_input_message_stream_open = False
    time.sleep(1.0)
    is_input_message_stream_open =True

    previous_inputs = []
    start = int(time.time())
    while is_input_message_stream_open:
        inputs, _ = get_inputs(controller_id)
        if inputs is not None and not np.array_equal(inputs, previous_inputs):
            previous_inputs = inputs 
            yield 'data: %s\n\n' % str(inputs).replace("'", "\"").replace("None", "0")
        else:
            # send an empty message after 45 seconds to avoid breaking the connection   
            end = int(time.time())
            if end - start >= 45:
                start = end
                yield 'data: []\n\n'
        # wait 50 ms for next iteration
        time.sleep(0.05)

    print("input message stream closed")


def counter_message_stream(controller_id):

    # stop all counter message streams before starting a new one
    global is_counter_message_stream_open
    is_counter_message_stream_open = False
    time.sleep(1.0)
    is_counter_message_stream_open =True

    previous_counters = []
    start = int(time.time())
    while is_counter_message_stream_open:
        counters, _ = get_counters(controller_id)
        if counters is not None and not np.array_equal(counters, previous_counters):
            previous_counters = counters
            yield 'data: %s\n\n' % str(counters).replace("'", "\"").replace("None", "0")
        else:
            # send an empty message after 45 seconds to avoid breaking the connection   
            end = int(time.time())
            if end - start >= 45:
                start = end
                yield 'data: []\n\n'
        # wait 50 ms for next iteration
        time.sleep(0.05)

    print("counter message stream closed")


def discover():
    response = Txt.get_instance().discover()
    return response, HTTPStatus.OK


def stop_controller():  # noqa: E501
    Txt.get_instance().stop()
    return None, HTTPStatus.OK


def start_interface_test():
    ftlock.lock('Test Interface')
    ftlock.notify(TxtState.INTERFACE_TEST_START.value)
    return None, HTTPStatus.OK


def stop_interface_test():
    stop_controller()
    ftlock.lock('Test Interface')
    ftlock.notify(TxtState.INTERFACE_TEST_STOP.value)
    ftlock.unlock()
    return None, HTTPStatus.OK
