import os
import shutil
import json
import ftlock
from http import HTTPStatus

from flask import jsonify
from txtapi.environment.environment import MANIFEST_FILENAME, PID_FILENAME, PROJECT_FILENAME
from txtapi.environment.environment import WORKSPACE_BASEPATH, DEBUG
from txtapi.models.file import File
from txtapi.models.workspace import Workspace
from txtapi.txt.txt_state import TxtState


def create_workspace(workspace_name):
    path = os.path.join(WORKSPACE_BASEPATH, workspace_name)
    if not os.path.exists(path):
        if DEBUG:
            print("Create directory %s" % (path))
        try:
            os.makedirs(path)
        except Exception as exception:
            if DEBUG:
                print("Creating directory %s failed due to %s" % (path, str(exception)))
            return exception, HTTPStatus.INTERNAL_SERVER_ERROR
    return build_workspace_response(workspace_name), HTTPStatus.CREATED


def get_workspace(workspace_name):
    path = os.path.join(WORKSPACE_BASEPATH, workspace_name)
    if os.path.exists(path):
        return build_workspace_response(workspace_name), HTTPStatus.OK
    else:
        return None, HTTPStatus.NOT_FOUND


def get_workspaces():
    workspaces = []
    for file in os.listdir(WORKSPACE_BASEPATH):
        path = os.path.join(WORKSPACE_BASEPATH, file)
        if os.path.isdir(path):
            workspaces.append(build_workspace_response(file))
    return workspaces, HTTPStatus.OK


def delete_file_of_workspace(workspace_name, filename):
    path = os.path.join(WORKSPACE_BASEPATH, workspace_name, filename)
    if os.path.exists(path):
        os.remove(path)
        return None, HTTPStatus.OK
    else:
        return None, HTTPStatus.NOT_FOUND


def delete_files_of_workspace(workspace_name):
    path = os.path.join(WORKSPACE_BASEPATH, workspace_name)
    if os.path.exists(path):
        for file in os.listdir(path):
            file_path = os.path.join(path, file)
            if os.path.isfile(file_path):
                os.remove(file_path)
            else:
                shutil.rmtree(file_path)
        return None, HTTPStatus.OK
    else:
        return None, HTTPStatus.NOT_FOUND


def delete_workspace(workspace_name):
    path = os.path.join(WORKSPACE_BASEPATH, workspace_name)
    if os.path.exists(path):
        shutil.rmtree(path)
        return None, HTTPStatus.OK
    else:
        return None, HTTPStatus.NOT_FOUND


def get_file_of_workspace(workspace_name, filename):
    path = os.path.join(WORKSPACE_BASEPATH, workspace_name, filename)
    if os.path.exists(path):
        return jsonify(File(filename, path)), HTTPStatus.OK
    else:
        return None, HTTPStatus.NOT_FOUND


def get_files_of_workspace(workspace_name):
    path = os.path.join(WORKSPACE_BASEPATH, workspace_name)
    if not os.path.exists(path):
        return None, HTTPStatus.NOT_FOUND
    files = []
    for file in os.listdir(path):
        files.append(File(file, os.path.join(path, file)))
    return files, HTTPStatus.OK


def upload_files_to_workspace(workspace_name, files):
    if ftlock.try_lock(workspace_name + '.py') == 0:
        ftlock.notify(TxtState.TRANSFERRING.value)
        ftlock.unlock()
    for file in files:
        if DEBUG:
            print(file.filename + ' uploaded\n\n')

        splitPath = file.filename.replace('../', '').split('/')
        fileName = splitPath[-1]
        path = '/'.join(splitPath[:-1])
        file_path = os.path.join(WORKSPACE_BASEPATH, workspace_name, path)
        
        if not os.path.exists(file_path):
            try:
                os.makedirs(file_path)
            except Exception as exception:
                return exception, HTTPStatus.INTERNAL_SERVER_ERROR
        
        file_path = os.path.join(file_path, fileName)
        file.save(file_path)
    if ftlock.try_lock(workspace_name + '.py') == 0:
        ftlock.notify(TxtState.STOP.value)
        ftlock.unlock()
    return HTTPStatus.CREATED


def get_base_path(workspace_name: str) -> str:
    return os.path.join(WORKSPACE_BASEPATH, workspace_name)


def get_file_path(workspace_name: str, relativepath: str) -> str:
    return os.path.join(get_base_path(workspace_name), relativepath)


def get_pid_file_path(workspace_name: str) -> str:
    return os.path.join(get_base_path(workspace_name), PID_FILENAME)


def get_project_file_path(workspace_name: str) -> str:
    return os.path.join(get_base_path(workspace_name), PROJECT_FILENAME)


def get_pid(workspace_name: str) -> str:
    try:
        pidFile = open(get_pid_file_path(workspace_name), "r")  # approot/.pid contains the pid
        pid = pidFile.read()
    except Exception as ex:
        raise Exception("unable to read .pid-file for workspace_name " + workspace_name + ". " + str(ex))
    if len(pid) == 0:
        raise Exception("unable to find process with workspace_name " + workspace_name)
    return pid

def build_workspace_response(name):
    uuid = None
    try:
        with open(get_project_file_path(name), "r") as project_file:
            project_dict = json.load(project_file)
            if 'uuid' in project_dict:
                uuid = project_dict['uuid']
    except: 
        pass
    path = os.path.join(WORKSPACE_BASEPATH, name)
    return Workspace(name, path, uuid)
